"""Flardl--download list of URLs from a list of federated servers."""

from .common import ALL  # noqa: F401
from .common import AVG  # noqa: F401
from .common import HIST  # noqa: F401
from .common import INDEX_KEY  # noqa: F401
from .common import MAX  # noqa: F401
from .common import MIN  # noqa: F401
from .common import NOBS  # noqa: F401
from .common import RAVG  # noqa: F401
from .common import SUM  # noqa: F401
from .common import VALUE  # noqa: F401
from .multidispatcher import MultiDispatcher  # noqa: F401
from .stream_stats import StreamStats  # noqa: F401
from .stream_stats import WorkerStat  # noqa: F401



"""Shared constants, parameters, and routines."""
from time import time
from typing import Optional
from typing import Union

# third-party imports
import numpy as np


# The following globals are also attribute names,
# don't change one without the other.
ALL = "all"
AVG = "avg"
HIST = "history"
INDEX_KEY = "idx"
MAX = "max"
MIN = "min"
NOBS = "n_obs"
RAVG = "r_avg"
SUM = "sum"
VALUE = "value"

# Pretty labels for substat names.
STAT_SUBLABELS = {
    VALUE: "",
    SUM: "total ",
    AVG: "average ",
    MIN: "min ",
    MAX: "max ",
    NOBS: "# ",
    HIST: "history ",
    RAVG: "rolling average ",
}
DEFAULT_ROUNDING = 2  # digits after decimal
DEFAULT_MAX_RETRIES = 0
TIME_ROUNDING = 1  # digits, milliseconds
RATE_ROUNDING = 1  # digits, inverse seconds
TIME_EPSILON = 0.01  # milliseconds
BYTES_TO_MEGABITS = 8.0 / 1024.0 / 1024.0
RANDOM_SEED = 87507
DEFAULT_ZIPF_EXPONENT = 1.5  # more divergent as it gets closer to 1
DEFAULT_ZIPF_SCALE = 1000
DEFAULT_ZIPF_MIN = 1024
NUMERIC_TYPE = Union[int, float]
OPTIONAL_NUMERIC = Optional[NUMERIC_TYPE]
OPTIONAL_NUMERIC_LIST = Union[OPTIONAL_NUMERIC, list[NUMERIC_TYPE]]
SIMPLE_TYPES = Union[int, float, bool, str, None]


class RandomValueGenerator:
    """Seeded, reproducible random-value generation."""

    def __init__(
        self,
        seed: int = RANDOM_SEED,
        zipf_minimum: int = DEFAULT_ZIPF_MIN,
        zipf_scale: int = DEFAULT_ZIPF_SCALE,
        zipf_exponent: float = DEFAULT_ZIPF_EXPONENT,
    ):
        """Init random value generator with seed."""
        self.rng = np.random.default_rng(seed=seed)
        self.zipf_minimum = zipf_minimum
        self.zipf_scale = zipf_scale
        self.zipf_exponent = zipf_exponent

    def get_wait_time(self, rate: float) -> float:
        """Given rate, return wait time from an exponential distribution."""
        return self.rng.exponential(1.0 / rate)

    def zipf_with_min(self) -> int:
        """Return a Zipf-law-distributed integer with minimum.

        This distribution approximately describes many file-size
        distributions arising from natural language and human-written
        code (though code standards discourage large files). It
        is a sensitive function of the exponent, with exponents
        near 1.0 generating a higher divergence. For exponents
        greater than 2, there is only a small likelihood of
        finding a 2-digit value in a sample of size 100; for
        an exponent of 1.1, most values will have multiple digits.
        The default value is chosen to give a reasonable range
        in a sample of a few thousand.

        Note that the first moment of this power-law
        distribution tends to increase with sample size.
        If used in a mock downloader as a sample size, this
        implies the mean per-file download rate goes down
        with the number of files downloaded because the chances
        of hitting a big file goes up.
        """
        return self.zipf_minimum + self.zipf_scale * self.rng.zipf(self.zipf_exponent)


class MillisecondTimer:
    """Give the time in milliseconds since initialization."""

    def __init__(self) -> None:
        """Init the start_time."""
        self.start_time = time()

    def time(self) -> float:
        """Return time from start in milliseconds."""
        return round((time() - self.start_time) * 1000.0, TIME_ROUNDING)



"""Work is dispatched to multiple workers and results collected via AnyIO streams."""
from __future__ import annotations

import sys
from typing import Any
from typing import Optional

# third-party imports
import anyio
import loguru
from loguru import logger as mylogger

from .common import DEFAULT_MAX_RETRIES
from .common import INDEX_KEY
from .common import SIMPLE_TYPES
from .common import MillisecondTimer
from .dict_to_indexed_list import NonStringIterable
from .dict_to_indexed_list import zip_dict_to_indexed_list
from .downloader import Downloader
from .downloader import MockDownloader
from .instrumented_streams import ArgumentStream
from .instrumented_streams import FailureStream
from .instrumented_streams import ResultStream
from .stream_stats import StreamStats


class MultiDispatcher:
    """Runs multiple single-site dispatchers sharing streams."""

    def __init__(  # noqa: C901
        self,
        all_worker_defs: list[dict[str, Any]],
        /,
        worker_list: list[str] | None = None,
        max_retries: int = DEFAULT_MAX_RETRIES,
        logger: loguru.Logger | None = None,
        quiet: bool = False,
        history_len: int = 0,
        output_dir: str | None = None,
        mock: bool = False,
    ) -> None:
        """Save list of dispatchers."""
        if logger is None:
            self._logger = mylogger
        else:
            self._logger = logger
        all_worker_names = [w["name"] for w in all_worker_defs]
        if worker_list is None:
            worker_defs = all_worker_defs
        else:
            worker_defs = []
            for worker_name in worker_list:
                try:
                    worker_idx = all_worker_names.index(worker_name)
                except ValueError:
                    self._logger.error(f"Worker name {worker_name} not found.")
                    sys.exit(1)
                worker_defs.append(all_worker_defs[worker_idx])
        self.workers = []
        if mock:
            worker_factory = MockDownloader
        else:
            worker_factory = Downloader
        for i, worker_def in enumerate(worker_defs):
            try:
                worker = worker_factory(
                    i, self._logger, output_dir, quiet, **worker_def
                )
            except Exception as e:
                self._logger.warning(
                    f"Worker {worker_def['name']} failed to initialize."
                )
                self._logger.warning(e)
                continue
            self.workers.append(worker)
        if len(self.workers) == 0:
            self._logger.error("No valid workers found.")
            sys.exit(1)
        self.max_retries = max_retries
        self.exception_counter: dict[int, int] = {}
        self.n_too_many_retries = 0
        self.n_exceptions = 0
        self.quiet = quiet
        self.queue_stats = StreamStats(all_worker_names, history_len=history_len)
        self._lock = anyio.Lock()
        self.inflight: dict[str, SIMPLE_TYPES] = {}
        self.timer = MillisecondTimer()

    async def run(
        self,
        args: (
            list[dict[str, SIMPLE_TYPES]] | dict[str, NonStringIterable | SIMPLE_TYPES]
        ),
    ):
        """Run the multidispatcher queue."""
        if isinstance(args, list):
            arg_list = args
        elif isinstance(args, dict):
            arg_list = zip_dict_to_indexed_list(args)
        arg_q = ArgumentStream(arg_list, self.inflight, self.timer)
        result_stream = ResultStream(self.inflight)
        failure_stream = FailureStream(self.inflight)

        async with anyio.create_task_group() as tg:
            for worker in self.workers:
                tg.start_soon(
                    self.dispatcher, worker, arg_q, result_stream, failure_stream
                )

        # Process results into pandas data frame in input order.
        results = result_stream.get_all()
        fails = failure_stream.get_all()
        stats = {
            "requests": len(arg_list),
            "downloaded": len(results),
            "failed": len(fails),
            "workers": len(self.workers),
        }
        return results, fails, stats

    async def dispatcher(  # noqa: C901
        self,
        worker,
        arg_q: ArgumentStream,
        result_q: ResultStream,
        failure_q: FailureStream,
    ):
        """Dispatch tasks to worker functions and handle exceptions."""
        while True:
            try:
                # Get a set of arguments from the queue.
                kwargs, worker_count = await arg_q.get(worker_name=worker.name)
            except anyio.WouldBlock:
                return
            # Do rate limiting, if a limiter is found in worker.
            try:
                await worker.limiter()
            except AttributeError:
                pass  # it's okay if worker didn't have a limiter
            # Do the work and handle any exceptions encountered.
            try:
                await worker.worker(result_q, worker_count, **kwargs)
            except worker.soft_exceptions as e:
                # Errors to be requeued by worker, unless too many
                idx = kwargs[INDEX_KEY]
                async with self._lock:
                    idx = kwargs[INDEX_KEY]
                    self.n_exceptions += 1
                    if idx not in self.exception_counter:
                        self.exception_counter[idx] = 1
                    else:
                        self.exception_counter[idx] += 1
                    n_exceptions = self.exception_counter[idx]
                if self.max_retries > 0 and n_exceptions >= self.max_retries:
                    await worker.hard_exception_handler(
                        idx, worker.name, worker_count, e, failure_q
                    )
                else:
                    await worker.soft_exception_handler(
                        kwargs, worker.name, worker_count, e, arg_q
                    )
            except worker.hard_exceptions as e:
                idx = kwargs[INDEX_KEY]
                await worker.hard_exception_handler(
                    idx, worker.name, worker_count, e, failure_q
                )
            except Exception as e:
                # unhandled errors go to unhandled exception handler
                idx = kwargs[INDEX_KEY]
                await worker.unhandled_exception_handler(idx, e)

    def main(self, arg_list: list[dict[str, SIMPLE_TYPES]], config: str = "production"):
        """Start the multidispatcher queue."""
        backend_options = {}
        if config == "production":
            backend = "asyncio"
            if sys.platform != "win32":
                backend_options = {"use_uvloop": True}
        elif config == "testing":
            backend = "asyncio"
            # asyncio.set_event_loop_policy(DeterministicEventLoopPolicy())
        elif config == "trio":
            backend = "trio"
        else:
            self._logger.error(f"Unknown configuration {config}")
            sys.exit(1)
        return anyio.run(
            self.run, arg_list, backend=backend, backend_options=backend_options
        )



"""Simple stream-associated statistical functions."""
from collections import UserDict
from collections import deque
from typing import Any
from typing import Optional
from typing import TypeVar
from typing import Union

from attrs import Attribute
from attrs import asdict
from attrs import define
from attrs import field

from .common import ALL
from .common import BYTES_TO_MEGABITS
from .common import DEFAULT_ROUNDING
from .common import HIST
from .common import MAX
from .common import NUMERIC_TYPE
from .common import RAVG
from .common import STAT_SUBLABELS
from .common import SUM
from .common import VALUE


# type definitions
OPTIONAL_NUMERIC = Optional[NUMERIC_TYPE]
OPTIONAL_NUMERIC_LIST = Union[OPTIONAL_NUMERIC, list[NUMERIC_TYPE]]

StreamStatsType = TypeVar("StreamStatsType", bound="StreamStats")
StatType = TypeVar("StatType", bound="Stat")
U = TypeVar("U")


def _nn(inst: Optional[U]) -> U:
    """Not-none helper to pass mypy."""
    assert inst is not None
    return inst


def _round(val: NUMERIC_TYPE, rounding: int) -> NUMERIC_TYPE:
    """Round with zero digits returning an int."""
    rounded_val = round(val, rounding)
    if rounding == 0:
        return int(rounded_val)
    return rounded_val


def _set_stat(instance: StatType, attrib: Attribute, val: NUMERIC_TYPE) -> NUMERIC_TYPE:
    """Round stat value and set derived quantities."""
    _unused = (attrib,)  # noqa: F841
    rounded_val = _round(val, instance._rounding)
    instance.n_obs += 1
    if instance.value is None:
        instance.min = rounded_val
        instance.max = rounded_val
        instance.sum = rounded_val
        instance.avg = rounded_val
    else:
        assert instance.min is not None
        instance.min = min(rounded_val, _nn(instance.min))
        instance.max = max(rounded_val, _nn(instance.max))
        instance.sum = _round(_nn(instance.sum) + rounded_val, instance._rounding)
        instance.avg = _round(_nn(instance.sum) / instance.n_obs, instance._rounding)
    instance._history.append(rounded_val)
    if instance._history_len > 0 and len(instance._history) == instance._history_len:
        instance.r_avg = _round(
            float(sum(instance._history)) / instance._history_len, instance._rounding
        )
    return rounded_val


@define
class Stat:
    """Class of derived statistics on numeric value."""

    _history_len: int = field(default=0, repr=False)
    _rounding: int = field(default=DEFAULT_ROUNDING, repr=False)
    value: OPTIONAL_NUMERIC = field(default=None, on_setattr=_set_stat)
    sum: OPTIONAL_NUMERIC = None
    n_obs: int = 0
    avg: OPTIONAL_NUMERIC = None
    min: OPTIONAL_NUMERIC = None
    max: OPTIONAL_NUMERIC = None
    _history: deque[NUMERIC_TYPE] = field(init=False, repr=False)
    r_avg: OPTIONAL_NUMERIC = None

    def __attrs_post_init__(self):
        """Initialize history after history length is set."""
        self._history = deque(maxlen=self._history_len)

    def get(self, key: str = VALUE) -> OPTIONAL_NUMERIC_LIST:
        """Return value of attribute."""
        if key is HIST:
            if self._history_len > 0 and len(self._history) == self._history_len:
                return list(self._history)
            else:
                return None
        return asdict(self, filter=lambda attr, value: not attr.name.startswith("_"))[
            key
        ]


class WorkerStat(UserDict):
    """Calculate stats on individual and all workers."""

    def __init__(
        self,
        label: str,
        workers: list[str],
        rounding: int = DEFAULT_ROUNDING,
        history_len: int = 0,
    ):
        """Initialize storage of stats."""
        self.label = label
        self.rounding = rounding
        self.history_len = history_len
        super().__init__()
        for worker in workers:
            self[worker] = Stat(rounding=self.rounding, history_len=self.history_len)

    def __repr__(self):
        """String of the overall stats."""
        return str(self[ALL])

    def set(
        self, value: NUMERIC_TYPE, worker: str = ALL, set_global: bool = True
    ) -> None:
        """Set value for a worker."""
        self[worker].value = value
        if worker is not ALL and set_global:
            self[ALL].value = value

    def get(
        self,
        key: str,
        /,
        default: Any = None,
        worker: str = ALL,
        scale: Optional[float] = None,
        rounding: Optional[int] = None,
    ) -> OPTIONAL_NUMERIC_LIST:
        """Get stat for a worker."""
        retval = self[worker].get(key)
        if scale is not None:
            retval *= scale
        if rounding is not None:
            retval = _round(retval, rounding)
        return retval


@define
class StatData:
    """Data class for WorkerStat initialization."""

    label: str
    rounding: int = DEFAULT_ROUNDING


@define
class ReportData:
    """Data class for report generation."""

    stat: str
    substat: str
    label: Optional[str] = None
    name: Optional[str] = None
    scale: Optional[float] = None
    rounding: Optional[int] = None


class StreamStats(UserDict):
    """Dictionary of per-worker queue stats with update calculations."""

    stat_data = {
        "retirement_t": StatData("retirement time, ms", rounding=2),
        "launch_t": StatData("launch time, ms", rounding=2),
        "service_t": StatData("service time, ns", rounding=2),
        "bytes": StatData("bytes downloaded", rounding=0),
        "dl_rate": StatData("per-file download rate, /s", rounding=1),
        "cum_rate": StatData("download rate, Mbit/s", rounding=0),
    }
    worker_stats = [
        ReportData(
            "retirement_t",
            VALUE,
            "Elapsed time, s",
            "elapsed_t",
            scale=1.0 / 1000.0,
            rounding=1,
        ),
        ReportData("dl_rate", MAX),
        ReportData(
            "bytes",
            SUM,
            scale=1 / 1024.0 / 1024.0,
            rounding=1,
            label="Total MB downloaded",
        ),
    ]
    file_stats = [
        ReportData("retirement_t", VALUE),
        ReportData("launch_t", VALUE),
        ReportData("service_t", VALUE),
        ReportData("bytes", VALUE),
    ]
    diagnostic_stats = [
        ReportData("dl_rate", RAVG),
    ]

    def __init__(self, workers: list[str], history_len: int):
        """Initialize dict of worker stats."""
        self.workers: list[str] = [ALL]
        if workers is not None:
            self.workers = workers + self.workers
        for report_list in [self.worker_stats, self.file_stats, self.diagnostic_stats]:
            for stat in report_list:
                if stat.name is None:
                    stat.name = stat.stat
                    if stat.substat != VALUE:
                        stat.name += "_" + stat.substat
                if stat.label is None:
                    stat.label = (
                        STAT_SUBLABELS[stat.substat] + self.stat_data[stat.stat].label
                    ).capitalize()
        super().__init__(
            {
                s: WorkerStat(
                    d.label,
                    workers=self.workers,
                    rounding=d.rounding,
                    history_len=history_len,
                )
                for s, d in self.stat_data.items()
            }
        )

    def update_stats(self, *args, worker: str = ALL) -> None:
        """Update using update methods in queue stats."""
        if len(args) > 0:
            input_dict = args[0].copy()
            pop_list = []
            for k, v in input_dict.items():
                if k not in self:
                    continue
                if not isinstance(v, WorkerStat):
                    self[k].set(v, worker=worker)
                    pop_list.append(k)
            [input_dict.pop(k) for k in pop_list]  # noqa: W0106
            super().update(input_dict)
        else:
            super().update(*args)
        self.calculate_updates(worker=worker)

    def globals(self) -> dict[str, OPTIONAL_NUMERIC]:
        """Return global stats."""
        ret_dict: dict[str, OPTIONAL_NUMERIC] = {}
        for key in self:
            if not self[key].is_global_stat:
                continue
            ret_dict[key] = self[key].get()
        return ret_dict

    def results(
        self,
        worker: str = ALL,
    ) -> dict[str, OPTIONAL_NUMERIC]:
        """Return per-result stats."""
        ret_dict: dict[str, OPTIONAL_NUMERIC] = {}
        for key in self:
            if not self[key].is_result_stat:
                continue
            ret_dict[key] = self[key].get(VALUE, worker=worker)
        return ret_dict

    def report_worker_stats(self) -> dict[str, dict[Optional[str], Any]]:
        """Return per-worker stats."""
        ret_dict = {}
        for worker in self.workers:
            ret_dict.update(
                {
                    worker: {
                        s.name: self[s.stat].get(
                            s.substat, worker=worker, scale=s.scale, rounding=s.rounding
                        )
                        for s in self.worker_stats
                    }
                }
            )
        return ret_dict

    def report_summary_stats(
        self, worker: str = ALL
    ) -> dict[Optional[str], OPTIONAL_NUMERIC]:
        """Return summary stats with nice labels for a worker."""
        return {
            s.label: self[s.stat].get(
                s.substat, worker=worker, scale=s.scale, rounding=s.rounding
            )
            for s in self.worker_stats
        }

    def report_file_stats(
        self, worker: str = ALL, diagnostics: bool = False
    ) -> dict[str, OPTIONAL_NUMERIC]:
        """Return file stats."""
        stat_list = self.file_stats
        if diagnostics:
            stat_list += self.diagnostic_stats
        return {
            _nn(s.name): self[s.stat].get(
                s.substat, worker=worker, scale=s.scale, rounding=s.rounding
            )
            for s in stat_list
        }

    def calculate_updates(
        self,
        worker: str = ALL,
    ):
        """Calculate all derived values."""
        # service_t
        try:
            self["service_t"].set(
                self["retirement_t"].get(VALUE, worker)
                - self["launch_t"].get(VALUE, worker),
                worker,
            )
        except TypeError:
            pass
        # dl_rate
        try:
            self["dl_rate"].set(
                self["bytes"].get(VALUE, worker)
                * 1000.0
                / 1024.0
                / 1024.0
                / self["service_t"].get(VALUE, worker),
                worker=worker,
                set_global=False,
            )
            # update global value
            self["dl_rate"].set(
                self["bytes"].get(VALUE)
                * 1000.0
                / 1024.0
                / 1024.0
                / self["service_t"].get(VALUE)
            )
        except TypeError:
            pass
        # cum_rate
        try:
            self["cum_rate"].set(
                self["bytes"].get(SUM, worker)
                * BYTES_TO_MEGABITS
                * 1000.0
                / self["retirement_t"].get(VALUE, worker),
                worker=worker,
                set_global=False,
            )
            # now update value for all workers
            self["cum_rate"].set(
                self["bytes"].get(SUM)
                * BYTES_TO_MEGABITS
                * 1000.0
                / self["retirement_t"].get(VALUE)
            )
        except TypeError:
            pass



"""Zip a dict of lists to an indexed list of dicts."""

import abc
from collections.abc import Iterable
from itertools import zip_longest
from typing import Union
from typing import cast

import _collections_abc as cabc

from .common import INDEX_KEY
from .common import SIMPLE_TYPES


class NonStringIterable(metaclass=abc.ABCMeta):
    """A class to find iterables that are not strings."""

    __slots__ = ()

    @abc.abstractmethod
    def __iter__(self):
        """Fake iteration."""
        while False:
            yield None

    @classmethod
    def __subclasshook__(cls, c):
        """Check if non-string iterable."""
        if cls is NonStringIterable:
            if issubclass(c, str):
                return False

            return cabc._check_methods(c, "__iter__")
        return NotImplemented


def zip_dict_to_indexed_list(
    arg_dict: dict[str, Union[NonStringIterable, SIMPLE_TYPES]]
) -> list[dict[str, SIMPLE_TYPES]]:
    """Zip on the longest non-string iterables, adding an index."""
    ret_list = []
    iterable_args = [k for k in arg_dict if isinstance(arg_dict[k], NonStringIterable)]
    idx = 0
    for iter_tuple in zip_longest(
        *[cast(Iterable, arg_dict[k]) for k in iterable_args]
    ):
        args: dict[str, SIMPLE_TYPES] = {INDEX_KEY: idx}
        for key in arg_dict.keys():
            if key in iterable_args:
                args[key] = iter_tuple[iterable_args.index(key)]
            else:
                args[key] = cast(SIMPLE_TYPES, arg_dict[key])
        idx += 1
        ret_list.append(args)
    return ret_list



"""Streams instrumented with depth and other stats."""

import math
from collections import Counter
from typing import Any
from typing import Union
from typing import cast

# third-party imports
import anyio

from .common import INDEX_KEY
from .common import RATE_ROUNDING
from .common import SIMPLE_TYPES
from .common import TIME_EPSILON
from .common import MillisecondTimer


LAUNCH_KEY = "launch_t"


def get_index_value(item: dict[str, SIMPLE_TYPES]) -> int:
    """Return value of index key."""
    return cast(int, item[INDEX_KEY])


class ArgumentStream:
    """A stream of dictionaries to be used as arguments."""

    def __init__(
        self,
        arg_list: list[dict[str, SIMPLE_TYPES]],
        in_process: dict[str, Any],
        timer: MillisecondTimer,
    ):
        """Initialize data structure for in-flight stats."""
        self.send_stream, self.receive_stream = anyio.create_memory_object_stream(
            max_buffer_size=math.inf
        )
        for args in arg_list:
            self.send_stream.send_nowait(args)
        self.n_args = len(arg_list)
        self.inflight = in_process
        self.timer = timer
        self.launch_rate = 0.0
        self.worker_counter: Counter[str] = Counter()
        self._lock = anyio.Lock()

    async def put(
        self,
        args,
        /,
        worker_name: Union[str, None] = None,
        worker_count: Union[int, None] = None,
    ):
        """Put on results queue and update stats."""
        worker_name = cast(str, worker_name)
        worker_count = cast(int, worker_count)
        async with self._lock:
            del self.inflight[worker_name][worker_count]
        await self.send_stream.send(args)

    async def get(self, /, worker_name: Union[str, None] = None, **kwargs):
        """Track de-queuing by worker."""
        worker_name = cast(str, worker_name)
        q_entry = self.receive_stream.receive_nowait(**kwargs)
        async with self._lock:
            self.worker_counter[worker_name] += 1
            worker_count = self.worker_counter[worker_name]
            if worker_name not in self.inflight:
                self.inflight[worker_name] = {}
            idx = q_entry[INDEX_KEY]
            launch_time = self.timer.time()
            self.launch_rate = round(
                idx * 1000.0 / (launch_time + TIME_EPSILON), RATE_ROUNDING
            )
            self.inflight[worker_name][worker_count] = {
                INDEX_KEY: idx,
                "queue_depth": len(self.inflight[worker_name]),
                LAUNCH_KEY: launch_time,
                "cum_launch_rate": self.launch_rate,
            }
        return q_entry, worker_count


class FailureStream:
    """Anyio stream to track failures."""

    launch_stats_out: list[str] = []

    def __init__(
        self,
        in_process: dict[str, Any],
    ) -> None:
        """Init stats for queue."""
        self.send_stream, self.receive_stream = anyio.create_memory_object_stream(
            max_buffer_size=math.inf
        )
        self.inflight = in_process
        self.count = 0
        self._lock = anyio.Lock()

    async def put(
        self,
        args,
        /,
        worker_name: Union[str, None] = None,
        worker_count: Union[int, None] = None,
    ):
        """Put on results queue and update stats."""
        worker_name = cast(str, worker_name)
        worker_count = cast(int, worker_count)
        launch_stats = self.inflight[worker_name][worker_count]
        for result_name in self.launch_stats_out:
            args[result_name] = launch_stats[result_name]
        async with self._lock:
            self.count += 1
            del self.inflight[worker_name][worker_count]
        await self.send_stream.send(args)

    def get_all(self) -> list[dict[str, SIMPLE_TYPES]]:
        """Return sorted list of stream contents."""
        stream_contents = []
        while True:
            try:
                stream_contents.append(self.receive_stream.receive_nowait())
            except anyio.WouldBlock:
                break
        return sorted(stream_contents, key=get_index_value)


class ResultStream(FailureStream):
    """Stream for results."""

    launch_stats_out = [LAUNCH_KEY]



"""Downloads as a MultiDispatcher worker class."""
from __future__ import annotations

import pathlib
import sys
from typing import Any

# third-party imports
import anyio
import httpx
import loguru

# module imports
from .common import INDEX_KEY
from .common import SIMPLE_TYPES
from .common import RandomValueGenerator
from .instrumented_streams import ArgumentStream
from .instrumented_streams import FailureStream
from .instrumented_streams import ResultStream


class StreamWorker:
    """Basic worker functions."""

    def __init__(
        self,
        worker_no: int,
        logger: loguru.Logger,
        output_dir: str | None,
        quiet: bool,
        /,
        name: str,
        bw_limit_mbps: float = 0.0,
        queue_depth: int = 0,
        timeout_factor: float = 0.0,
        **kwargs,
    ):
        """Positional=common across workers, keyworded=individual."""
        self.worker_no = worker_no
        self.output_dir = output_dir
        self._logger = logger
        self.output_dir = output_dir
        if output_dir is not None:
            pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)
        self.quiet = quiet
        # keyworded parameters
        self.name = name

        self.bw_limit = bw_limit_mbps
        self.queue_depth = queue_depth
        self.timeout_factor = timeout_factor
        # initialize internal parameters
        self.work_qty_name = "bytes"
        self.n_soft_fails = 0
        self.n_hard_fails = 0
        self.hard_exceptions: tuple[()] | tuple[type[BaseException]] = ()
        self.soft_exceptions: tuple[()] | tuple[type[BaseException]] = ()
        self._lock = anyio.Lock()
        self._limiter_delay = RandomValueGenerator().get_wait_time

    async def limiter(self):
        """Fake rate-limiting via sleep."""
        await anyio.sleep(self._limiter_delay(self.launch_rate))

    async def add_result(
        self,
        data: bytes | str,
        filename: str,
        idx: int,
        worker_count: int,
        result_q: ResultStream,
        /,
        **kwargs: dict[str, SIMPLE_TYPES],
    ):
        """Put dictionary of results on ouput queue."""
        work_qty = len(data)
        if self.output_dir is not None:
            out_file_str = self.output_dir + "/" + filename
            if isinstance(data, str):
                async with await anyio.open_file(out_file_str, "w") as fp:
                    await fp.write(data)
            else:
                out_file_path = anyio.Path(out_file_str)
                await out_file_path.write_bytes(data)
        results = {
            INDEX_KEY: idx,
            "worker": self.name,
            self.work_qty_name: work_qty,
        }
        results.update(kwargs)
        await result_q.put(results, worker_name=self.name, worker_count=worker_count)

    async def hard_exception_handler(
        self,
        index: int,
        worker_name: str,
        worker_count: int,
        error: Exception,
        failure_q: FailureStream,
    ):
        """Handle exceptions that re-queue arguments as failed."""
        if error.__class__ in self.soft_exceptions:
            message = repr(error)
            error_name = "TooManyRetries"
        else:
            message = str(error)
            error_name = error.__class__.__name__
        if not self.quiet:
            self._logger.error(f"{error_name}: {message}")
        failure_entry = {
            INDEX_KEY: index,
            "worker": worker_name,
            "error": error_name,
            "message": message,
        }
        await failure_q.put(
            failure_entry, worker_name=worker_name, worker_count=worker_count
        )

    async def unhandled_exception_handler(self, index: int, error: Exception):
        """Handle unhandled exceptions."""
        self._logger.error(error)
        await self._logger.complete()
        sys.exit(1)

    async def soft_exception_handler(
        self,
        kwargs: dict[str, Any],
        worker_name: str,
        worker_count: int,
        error: Exception,
        arg_q: ArgumentStream,
    ):
        """Handle exceptions that re-try arguments."""
        if not self.quiet:
            self._logger.warning(error)
        await arg_q.put(kwargs, worker_name=worker_name, worker_count=worker_count)


class MockDownloader(StreamWorker):
    """Demonstrates multi-dispatch operation with logging."""

    SOFT_FAILS = [2, 4]
    RESCUE_SOFT_FAILS = (4,)
    HARD_FAILS = (
        6,
        9,
    )
    LAUNCH_RETIREMENT_RATIO = 1.0
    LAUNCH_RATE_MAX = 100.0
    DL_RATE = 10000  # chunks/sec
    DL_CHUNK_SIZE = 1500  # bytes per chunk (packet)
    TIME_ROUND = 4

    def __init__(self, *args, **kwargs):
        """Init with id number."""
        super().__init__(*args, **kwargs)
        self.hard_exceptions: tuple[()] | tuple[type[BaseException]] = (ValueError,)
        self.soft_exceptions: tuple[()] | tuple[type[BaseException]] = (
            ConnectionError,
        )
        self.launch_rate = self.LAUNCH_RATE_MAX / (self.worker_no + 1.0)
        self.retirement_rate = self.launch_rate / self.LAUNCH_RETIREMENT_RATIO
        self._simulated_bytes = RandomValueGenerator().zipf_with_min
        self._simulated_dl_time = RandomValueGenerator().get_wait_time

    async def worker(
        self,
        result_q: ResultStream,
        worker_count: int,
        /,
        idx: int,
        code: str | None = None,
        file_type: str | None = None,
    ):
        """Do a work unit."""
        if idx in self.SOFT_FAILS:
            async with self._lock:
                self.n_soft_fails += 1
                if idx in self.RESCUE_SOFT_FAILS:
                    self.SOFT_FAILS.remove(idx)
            raise ConnectionError(f"{self.name} aborted job {idx} (expected).")
        elif idx in self.HARD_FAILS:
            async with self._lock:
                self.n_hard_fails += 1
            raise ValueError(f"Job {idx} failed on {self.name} (expected).")
        elif not self.quiet:
            self._logger.info(f"{self.name} working on job {idx}...")
        # create simulated output
        n_dl_bytes = self._simulated_bytes()
        dl_data = "a" * n_dl_bytes
        filename = str(code) + "." + str(file_type)
        # simulate download time with a sleep
        latency = self._simulated_dl_time(self.retirement_rate)
        receive_time = int(n_dl_bytes / self.DL_CHUNK_SIZE) / self.DL_RATE
        dl_time = round(latency + receive_time, self.TIME_ROUND)
        await anyio.sleep(dl_time)
        out_filename = filename
        await self.add_result(dl_data, out_filename, idx, worker_count, result_q)


class Downloader(StreamWorker):
    """Demonstrates multi-dispatch operation with logging."""

    LAUNCH_RATE_MAX = 100.0
    TIME_ROUND = 4

    def __init__(
        self,
        *args,
        server: str = "",
        dir: str = "",
        transport: str = "https",
        transport_ver: str = "1",
        **kwargs,
    ):
        """Init with id number."""
        super().__init__(*args, **kwargs)
        self.hard_exceptions: tuple[()] | tuple[type[BaseException]] = (ValueError,)
        self.soft_exceptions: tuple[()] | tuple[type[BaseException]] = (
            ConnectionError,
        )
        self.launch_rate = self.LAUNCH_RATE_MAX
        self.base_url = transport + "://" + server + "/"
        if dir != "":
            self.base_url += dir + "/"
        if transport_ver == "2":
            self.http2 = True
        else:
            self.http2 = False
        self.client = httpx.AsyncClient(base_url=self.base_url, http2=self.http2)

    async def worker(
        self,
        result_q: ResultStream,
        worker_count: int,
        /,
        idx: int,
        path: str,
        out_filename: str,
    ):
        """Download a file."""
        if not self.quiet:
            self._logger.info(
                f"Downloading {self.base_url}{path} to file {out_filename}"
            )
        response = await self.client.get(path)
        dl_data = response.content
        await self.add_result(dl_data, out_filename, idx, worker_count, result_q)
